r"""Contain DataFrame ingestors."""

from __future__ import annotations

__all__ = [
    "BaseIngestor",
    "CacheIngestor",
    "ClickHouseArrowIngestor",
    "CsvIngestor",
    "Ingestor",
    "ParquetIngestor",
    "TransformIngestor",
    "is_ingestor_config",
    "setup_ingestor",
]

from grizz.ingestor.base import BaseIngestor, is_ingestor_config, setup_ingestor
from grizz.ingestor.cache import CacheIngestor
from grizz.ingestor.clickhouse import ClickHouseArrowIngestor
from grizz.ingestor.csv import CsvIngestor
from grizz.ingestor.parquet import ParquetIngestor
from grizz.ingestor.transform import TransformIngestor
from grizz.ingestor.vanilla import Ingestor



r"""Contain transformers to replace values."""

from __future__ import annotations

__all__ = ["ReplaceStrictTransformer", "ReplaceTransformer"]

import logging
from typing import Any

import polars as pl
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseIn1Out1Transformer
from grizz.utils.format import str_kwargs

logger = logging.getLogger(__name__)


class ReplaceTransformer(BaseIn1Out1Transformer):
    r"""Replace the values in a column by the values in a mapping.

    Args:
        in_col: The input column name.
        out_col: The output column name.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments to pass to ``replace``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Replace
    >>> transformer = Replace(in_col="old", out_col="new", old={"a": 1, "b": 2, "c": 3})
    >>> transformer
    ReplaceTransformer(in_col='old', out_col='new', exist_policy='raise', missing_policy='raise', old={'a': 1, 'b': 2, 'c': 3})
    >>> frame = pl.DataFrame({"old": ["a", "b", "c", "d", "e"]})
    >>> frame
    shape: (5, 1)
    ┌─────┐
    │ old │
    │ --- │
    │ str │
    ╞═════╡
    │ a   │
    │ b   │
    │ c   │
    │ d   │
    │ e   │
    └─────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬─────┐
    │ old ┆ new │
    │ --- ┆ --- │
    │ str ┆ str │
    ╞═════╪═════╡
    │ a   ┆ 1   │
    │ b   ┆ 2   │
    │ c   ┆ 3   │
    │ d   ┆ d   │
    │ e   ┆ e   │
    └─────┴─────┘
    >>> transformer = Replace(
    ...     in_col="old",
    ...     out_col="new",
    ...     old={"a": 1, "b": 2, "c": 3},
    ...     default=None,
    ... )
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬──────┐
    │ old ┆ new  │
    │ --- ┆ ---  │
    │ str ┆ i64  │
    ╞═════╪══════╡
    │ a   ┆ 1    │
    │ b   ┆ 2    │
    │ c   ┆ 3    │
    │ d   ┆ null │
    │ e   ┆ null │
    └─────┴──────┘

    ```
    """

    def __init__(
        self,
        in_col: str,
        out_col: str,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            in_col=in_col,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "in_col": self._in_col,
                "out_col": self._out_col,
                "exist_policy": self._exist_policy,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Replacing values from column {self._in_col!r} and "
            f"saving output in {self._out_col!r} ..."
        )
        return frame.with_columns(pl.col(self._in_col).replace(**self._kwargs).alias(self._out_col))


class ReplaceStrictTransformer(BaseIn1Out1Transformer):
    r"""Replace the values in a column by the values in a mapping.

    Args:
        in_col: The input column name.
        out_col: The output column name.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments to pass to ``replace``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import ReplaceStrict
    >>> transformer = ReplaceStrict(
    ...     in_col="old", out_col="new", old={"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}
    ... )
    >>> transformer
    ReplaceStrictTransformer(in_col='old', out_col='new', exist_policy='raise', missing_policy='raise', old={'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})
    >>> frame = pl.DataFrame({"old": ["a", "b", "c", "d", "e"]})
    >>> frame
    shape: (5, 1)
    ┌─────┐
    │ old │
    │ --- │
    │ str │
    ╞═════╡
    │ a   │
    │ b   │
    │ c   │
    │ d   │
    │ e   │
    └─────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬─────┐
    │ old ┆ new │
    │ --- ┆ --- │
    │ str ┆ i64 │
    ╞═════╪═════╡
    │ a   ┆ 1   │
    │ b   ┆ 2   │
    │ c   ┆ 3   │
    │ d   ┆ 4   │
    │ e   ┆ 5   │
    └─────┴─────┘
    >>> transformer = ReplaceStrict(
    ...     in_col="old",
    ...     out_col="new",
    ...     old={"a": 1, "b": 2, "c": 3},
    ...     default=None,
    ... )
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬──────┐
    │ old ┆ new  │
    │ --- ┆ ---  │
    │ str ┆ i64  │
    ╞═════╪══════╡
    │ a   ┆ 1    │
    │ b   ┆ 2    │
    │ c   ┆ 3    │
    │ d   ┆ null │
    │ e   ┆ null │
    └─────┴──────┘

    ```
    """

    def __init__(
        self,
        in_col: str,
        out_col: str,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            in_col=in_col,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "in_col": self._in_col,
                "out_col": self._out_col,
                "exist_policy": self._exist_policy,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Replacing values from column {self._in_col!r} and "
            f"saving output in {self._out_col!r} ..."
        )
        return frame.with_columns(
            pl.col(self._in_col).replace_strict(**self._kwargs).alias(self._out_col)
        )



r"""Contain utility functions for counting."""

from __future__ import annotations

__all__ = ["compute_nunique", "compute_temporal_count", "compute_temporal_value_counts"]

from unittest.mock import Mock

import polars as pl
from coola.utils import is_numpy_available
from coola.utils.imports import check_numpy

from grizz.utils.sorting import mixed_typed_sort
from grizz.utils.temporal import to_step_names

if is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    np = Mock()


def compute_nunique(frame: pl.DataFrame) -> np.ndarray:
    r"""Return the number of unique values in each column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        An array with the number of unique values in each column.
            The shape of the array is the number of columns.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.count import compute_nunique
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"int": pl.Int64, "float": pl.Float64, "str": pl.String},
    ... )
    >>> count = compute_nunique(frame)
    >>> count
    array([3, 4, 3])

    ```
    """
    check_numpy()
    if (ncols := frame.shape[1]) == 0:
        return np.zeros(ncols, dtype=np.int64)
    return frame.select(pl.all().n_unique()).to_numpy()[0].astype(np.int64)


def compute_temporal_count(
    frame: pl.DataFrame,
    temporal_column: str,
    period: str,
) -> tuple[np.ndarray, list[str]]:
    r"""Prepare the data to create the figure and table.

    Args:
        frame: The DataFrame to analyze.
        temporal_column: The temporal column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A tuple with the counts and the temporal steps.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.count import compute_temporal_count
    >>> counts, steps = compute_temporal_count(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [None, float("nan"), 0.0, 1.0, 4.2, 42.0],
    ...             "col2": [None, 1, 0, None, 2, 3],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col1": pl.Float64,
    ...             "col2": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     temporal_column="datetime",
    ...     period="1mo",
    ... )
    >>> counts
    array([3, 1, 1, 1])
    >>> steps
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    check_numpy()
    if frame.is_empty():
        return np.array([], dtype=np.int64), []

    groups = (
        frame.select(pl.col(temporal_column).alias("datetime"), pl.lit(1).alias("count"))
        .sort("datetime")
        .group_by_dynamic("datetime", every=period)
    )
    steps = to_step_names(groups=groups, period=period)
    counts = groups.agg(pl.col("count").sum())["count"].to_numpy().astype(np.int64)
    return counts, steps


def compute_temporal_value_counts(
    frame: pl.DataFrame,
    column: str,
    temporal_column: str,
    period: str,
    drop_nulls: bool = False,
) -> tuple[np.ndarray, list[str], list[str]]:
    r"""Compute the value counts for temporal windows of a given column.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze the temporal value counts.
        temporal_column: The temporal column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        drop_nulls: If ``True``, the null values are ignored.

    Returns:
        A tuple with 3 items. The first item is a 2-d array that
            indicates the number of occurrences for each value and
            time step. The first dimension represents the value and
            the second dimension represents the steps. The second item
            is the list of time steps. The third item is the list of
            string representation of the values.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.count import compute_temporal_value_counts
    >>> counts, steps, values = compute_temporal_value_counts(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [None, 1.0, 0.0, 1.0, 4.2, 42.0],
    ...             "col2": [None, 1, 0, None, 2, 3],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col1": pl.Float64,
    ...             "col2": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col1",
    ...     temporal_column="datetime",
    ...     period="1mo",
    ... )
    >>> counts
    array([[1, 0, 0, 0],
           [1, 1, 0, 0],
           [0, 0, 1, 0],
           [0, 0, 0, 1],
           [1, 0, 0, 0]])
    >>> steps
    ['2020-01', '2020-02', '2020-03', '2020-04']
    >>> values
    ['0.0', '1.0', '4.2', '42.0', 'null']

    ```
    """
    check_numpy()
    if frame.is_empty():
        return np.zeros((0, 0), dtype=np.int64), [], []

    frame = frame.select(
        pl.col(temporal_column).alias("__datetime__"), pl.col(column).alias("value")
    )
    if drop_nulls:
        frame = frame.drop_nulls()

    groups = frame.sort(["__datetime__", "value"]).group_by_dynamic("__datetime__", every=period)
    steps = to_step_names(groups=groups, period="1mo")
    frame_counts = (
        groups.agg(pl.col("value").value_counts())
        .explode("value")
        .unnest("value")
        .pivot(on="value", index="__datetime__", values="count")
        .drop("__datetime__")
    )
    frame_counts = frame_counts.select(mixed_typed_sort(frame_counts.columns))
    counts = frame_counts.fill_null(0.0).to_numpy().astype(np.int64).transpose()
    return counts, steps, list(frame_counts.columns)



r"""Contain ``polars.DataFrame`` transformers to convert some columns to
a new data type."""

from __future__ import annotations

__all__ = ["ToDatetimeTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs

from grizz.transformer.columns import BaseInNTransformer

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class ToDatetimeTransformer(BaseInNTransformer):
    r"""Implement a transformer to convert some columns to a
    ``polars.Datetime`` type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        format: Format to use for conversion. Refer to the
            [chrono crate documentation](https://docs.rs/chrono/latest/chrono/format/strftime/index.html)
            for the full specification.
            Example: ``"%Y-%m-%d %H:%M:%S"``.
            If set to ``None`` (default), the format is inferred from
            the data.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``to_datetime``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import ToDatetime
    >>> transformer = ToDatetime(columns=["col1"])
    >>> transformer
    ToDatetimeTransformer(columns=('col1',), exclude_columns=(), missing_policy='raise', format=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [
    ...             "2020-01-01 01:01:01",
    ...             "2020-01-01 02:02:02",
    ...             "2020-01-01 12:00:01",
    ...             "2020-01-01 18:18:18",
    ...             "2020-01-01 23:59:59",
    ...         ],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [
    ...             "2020-01-01 11:11:11",
    ...             "2020-02-01 12:12:12",
    ...             "2020-03-01 13:13:13",
    ...             "2020-04-01 08:08:08",
    ...             "2020-05-01 23:59:59",
    ...         ],
    ...     },
    ... )
    >>> frame
    shape: (5, 3)
    ┌─────────────────────┬──────┬─────────────────────┐
    │ col1                ┆ col2 ┆ col3                │
    │ ---                 ┆ ---  ┆ ---                 │
    │ str                 ┆ str  ┆ str                 │
    ╞═════════════════════╪══════╪═════════════════════╡
    │ 2020-01-01 01:01:01 ┆ 1    ┆ 2020-01-01 11:11:11 │
    │ 2020-01-01 02:02:02 ┆ 2    ┆ 2020-02-01 12:12:12 │
    │ 2020-01-01 12:00:01 ┆ 3    ┆ 2020-03-01 13:13:13 │
    │ 2020-01-01 18:18:18 ┆ 4    ┆ 2020-04-01 08:08:08 │
    │ 2020-01-01 23:59:59 ┆ 5    ┆ 2020-05-01 23:59:59 │
    └─────────────────────┴──────┴─────────────────────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌─────────────────────┬──────┬─────────────────────┐
    │ col1                ┆ col2 ┆ col3                │
    │ ---                 ┆ ---  ┆ ---                 │
    │ datetime[μs]        ┆ str  ┆ str                 │
    ╞═════════════════════╪══════╪═════════════════════╡
    │ 2020-01-01 01:01:01 ┆ 1    ┆ 2020-01-01 11:11:11 │
    │ 2020-01-01 02:02:02 ┆ 2    ┆ 2020-02-01 12:12:12 │
    │ 2020-01-01 12:00:01 ┆ 3    ┆ 2020-03-01 13:13:13 │
    │ 2020-01-01 18:18:18 ┆ 4    ┆ 2020-04-01 08:08:08 │
    │ 2020-01-01 23:59:59 ┆ 5    ┆ 2020-05-01 23:59:59 │
    └─────────────────────┴──────┴─────────────────────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        format: str | None = None,  # noqa: A002
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._format = format
        self._kwargs = kwargs

    def get_args(self) -> dict:
        return super().get_args() | {"format": self._format} | self._kwargs

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Converting {len(self.find_columns(frame)):,} columns to datetime "
            f"| {self._format!r} ..."
        )
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select(
                (cs.by_name(columns) & cs.string()).str.to_datetime(self._format, **self._kwargs)
            )
        )



r"""Contain transformers to concatenate columns."""

from __future__ import annotations

__all__ = ["ConcatColumnsTransformer"]

import logging

import polars as pl
import polars.selectors as cs

from grizz.transformer.columns import BaseInNOut1Transformer

logger = logging.getLogger(__name__)


class ConcatColumnsTransformer(BaseInNOut1Transformer):
    r"""Implement a transformer to concatenate columns into a new column.

    Args:
        columns: The columns to concatenate. The columns should have
            the same type or compatible types. If ``None``,
            it processes all the columns.
        out_col: The output column.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import ConcatColumns
    >>> transformer = ConcatColumns(columns=["col1", "col2", "col3"], out_col="col")
    >>> transformer
    ConcatColumnsTransformer(columns=('col1', 'col2', 'col3'), out_col='col', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [11, 12, 13, 14, 15],
    ...         "col2": [21, 22, 23, 24, 25],
    ...         "col3": [31, 32, 33, 34, 35],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 11   ┆ 21   ┆ 31   ┆ a    │
    │ 12   ┆ 22   ┆ 32   ┆ b    │
    │ 13   ┆ 23   ┆ 33   ┆ c    │
    │ 14   ┆ 24   ┆ 34   ┆ d    │
    │ 15   ┆ 25   ┆ 35   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 5)
    ┌──────┬──────┬──────┬──────┬──────────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col          │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---          │
    │ i64  ┆ i64  ┆ i64  ┆ str  ┆ list[i64]    │
    ╞══════╪══════╪══════╪══════╪══════════════╡
    │ 11   ┆ 21   ┆ 31   ┆ a    ┆ [11, 21, 31] │
    │ 12   ┆ 22   ┆ 32   ┆ b    ┆ [12, 22, 32] │
    │ 13   ┆ 23   ┆ 33   ┆ c    ┆ [13, 23, 33] │
    │ 14   ┆ 24   ┆ 34   ┆ d    ┆ [14, 24, 34] │
    │ 15   ┆ 25   ┆ 35   ┆ e    ┆ [15, 25, 35] │
    └──────┴──────┴──────┴──────┴──────────────┘


    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Concatenating {len(self.find_columns(frame)):,} columns to {self._out_col!r} ..."
        )
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select(pl.concat_list(cs.by_name(columns).alias(self._out_col)))
        )



r"""Contain ``polars.DataFrame`` transformers to compute difference."""

from __future__ import annotations

__all__ = ["DiffTransformer", "TimeDiffTransformer"]

import logging
from typing import TYPE_CHECKING

import polars as pl
from coola.utils.format import repr_mapping_line

from grizz.transformer.base import BaseTransformer
from grizz.transformer.columns import BaseIn1Out1Transformer

if TYPE_CHECKING:
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class DiffTransformer(BaseIn1Out1Transformer):
    r"""Implement a transformer to compute the first discrete difference
    between shifted items.

    Args:
        in_col: The input column name.
        out_col: The output column name.
        shift: The number of slots to shift.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Diff
    >>> transformer = Diff(in_col="col1", out_col="diff")
    >>> transformer
    DiffTransformer(in_col='col1', out_col='diff', exist_policy='raise', missing_policy='raise', shift=1)
    >>> frame = pl.DataFrame({"col1": [1, 2, 3, 4, 5], "col2": ["a", "b", "c", "d", "e"]})
    >>> frame
    shape: (5, 2)
    ┌──────┬──────┐
    │ col1 ┆ col2 │
    │ ---  ┆ ---  │
    │ i64  ┆ str  │
    ╞══════╪══════╡
    │ 1    ┆ a    │
    │ 2    ┆ b    │
    │ 3    ┆ c    │
    │ 4    ┆ d    │
    │ 5    ┆ e    │
    └──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ diff │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ a    ┆ null │
    │ 2    ┆ b    ┆ 1    │
    │ 3    ┆ c    ┆ 1    │
    │ 4    ┆ d    ┆ 1    │
    │ 5    ┆ e    ┆ 1    │
    └──────┴──────┴──────┘

    ```
    """

    def __init__(
        self,
        in_col: str,
        out_col: str,
        shift: int = 1,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            in_col=in_col,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )
        self._shift = shift

    def get_args(self) -> dict:
        return super().get_args() | {"shift": self._shift}

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Computing the first discrete difference between shifted items | "
            f"in_col={self._in_col!r} | out_col={self._out_col!r} | shift={self._shift}"
        )
        return frame.with_columns(
            frame.select(pl.col(self._in_col).diff(n=self._shift).alias(self._out_col))
        )


class TimeDiffTransformer(BaseTransformer):
    r"""Implement a transformer to compute the time difference between
    consecutive time steps.

    Args:
        group_cols: The columns used to generate the group for each
            sequence.
        time_col: The input time column name.
        time_diff_col: The output time difference column name.
        shift: The number of slots to shift.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import TimeDiff
    >>> transformer = TimeDiff(group_cols=["col"], time_col="time", time_diff_col="diff")
    >>> transformer
    TimeDiffTransformer(group_cols=['col'], time_col='time', time_diff_col='diff', shift=1)
    >>> frame = pl.DataFrame({"col": ["a", "b", "a", "a", "b"], "time": [1, 2, 3, 4, 5]})
    >>> frame
    shape: (5, 2)
    ┌─────┬──────┐
    │ col ┆ time │
    │ --- ┆ ---  │
    │ str ┆ i64  │
    ╞═════╪══════╡
    │ a   ┆ 1    │
    │ b   ┆ 2    │
    │ a   ┆ 3    │
    │ a   ┆ 4    │
    │ b   ┆ 5    │
    └─────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌─────┬──────┬──────┐
    │ col ┆ time ┆ diff │
    │ --- ┆ ---  ┆ ---  │
    │ str ┆ i64  ┆ i64  │
    ╞═════╪══════╪══════╡
    │ a   ┆ 1    ┆ 0    │
    │ a   ┆ 3    ┆ 2    │
    │ a   ┆ 4    ┆ 1    │
    │ b   ┆ 2    ┆ 0    │
    │ b   ┆ 5    ┆ 3    │
    └─────┴──────┴──────┘

    ```
    """

    def __init__(
        self, group_cols: Sequence[str], time_col: str, time_diff_col: str, shift: int = 1
    ) -> None:
        self._group_cols = list(group_cols)
        self._time_col = time_col
        self._time_diff_col = time_diff_col
        self._shift = shift

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "group_cols": self._group_cols,
                "time_col": self._time_col,
                "time_diff_col": self._time_diff_col,
                "shift": self._shift,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def fit_transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self.fit(frame)
        return self.transform(frame)

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Computing the time difference between consecutive time steps | "
            f"group_cols={self._group_cols} | time_col={self._time_col!r} | "
            f"time_diff_col={self._time_diff_col!r} | shift={self._shift}"
        )
        frame = frame.sort(by=[*self._group_cols, self._time_col])
        return frame.with_columns(
            frame.group_by(self._group_cols)
            .agg(pl.col(self._time_col).diff(n=1).replace({None: 0}).alias(self._time_diff_col))
            .sort(by=self._group_cols)
            .select(pl.col(self._time_diff_col).explode())
        )



r"""Contain utility functions to sort values from multiple types."""

from __future__ import annotations

__all__ = ["mixed_typed_sort"]

from collections import defaultdict
from typing import TYPE_CHECKING

from grizz.utils.nan import sortnan

if TYPE_CHECKING:
    from collections.abc import Iterable


def mixed_typed_sort(iterable: Iterable, /, *, reverse: bool = False) -> list:
    r"""Return a new list containing all items from the iterable sorted
    in ascending order.

    This function is an extension of the built-in ``sorted`` function
    that works on a list with multiple types. There is no global order
    for all types, so the items are sorted only by type. For example,
    if a list has string and float values, the string values are
    sorted together and the float values are sorted together.
    Each type must implement the python sorting interface.
    The types are sorted by alphabetical order, so in the previous
    example, the float values are before the string values in the
    sorted output list. This function uses ``sortnan`` to sort
    numerical values, so it is possible to sort a list with NaNs.

    Args:
        iterable: The data to sort.
        reverse: If set to ``True``, then the list elements are sorted
            as if each comparison were reversed.

    Returns:
        The sorted data.

    Example usage:

    ```pycon

    >>> from grizz.utils.sorting import mixed_typed_sort
    >>> x = [1, "c", "a", "b", 4, -2]
    >>> mixed_typed_sort(x)
    [-2, 1, 4, 'a', 'b', 'c']
    >>> mixed_typed_sort(x, reverse=True)
    [4, 1, -2, 'c', 'b', 'a']

    ```
    """
    typed_data = defaultdict(list)
    for v in iterable:
        t = type(v)
        if t in {float, int, bool}:  # group the numeric types
            t = float
        typed_data[t].append(v)
    output = []
    # There is no global order defined between the types so we use the string representations
    # to sort the types.
    typed_data = dict(sorted(typed_data.items(), key=lambda x: x[0].__qualname__))
    for typ, values in typed_data.items():
        sort_fn = sortnan if typ is float else sorted
        output.extend(sort_fn(values, reverse=reverse))
    return output



r"""Contain ``polars.DataFrame`` transformers to convert some columns to
a new data type."""

from __future__ import annotations

__all__ = [
    "CastTransformer",
    "CategoricalCastTransformer",
    "DecimalCastTransformer",
    "FloatCastTransformer",
    "IntegerCastTransformer",
]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseIn1Out1Transformer, BaseInNTransformer
from grizz.utils.format import str_size_diff

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class CastTransformer(BaseInNTransformer):
    r"""Implement a transformer to convert some columns to a new data
    type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        dtype: The target data type.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``cast``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Cast
    >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
    >>> transformer
    CastTransformer(columns=('col1', 'col3'), dtype=Int32, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i32  ┆ str  ┆ i32  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        dtype: type[pl.DataType],
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns, exclude_columns=exclude_columns, missing_policy=missing_policy
        )
        self._dtype = dtype
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    def get_args(self) -> dict:
        return {
            "columns": self._columns,
            "dtype": self._dtype,
            "exclude_columns": self._exclude_columns,
            "missing_policy": self._missing_policy,
        } | self._kwargs

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Casting {len(self.find_columns(frame)):,} columns to {self._dtype}...")
        columns = self.find_common_columns(frame)
        out = self._transform_frame(frame, columns)
        logger.info(str_size_diff(orig=frame.estimated_size(), final=out.estimated_size()))
        return out

    def _transform_frame(self, frame: pl.DataFrame, columns: Sequence[str]) -> pl.DataFrame:
        return frame.with_columns(
            frame.select(cs.by_name(columns).cast(self._dtype, **self._kwargs))
        )


class DecimalCastTransformer(CastTransformer):
    r"""Implement a transformer to convert columns of type decimal to a
    new data type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        dtype: The target data type.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``cast``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import DecimalCast
    >>> transformer = DecimalCast(columns=["col1", "col2"], dtype=pl.Float32)
    >>> transformer
    DecimalCastTransformer(columns=('col1', 'col2'), dtype=Float32, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...         "col3": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     },
    ...     schema={
    ...         "col1": pl.Int64,
    ...         "col2": pl.Decimal,
    ...         "col3": pl.Decimal,
    ...         "col4": pl.String,
    ...     },
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────────────┬──────────────┬──────┐
    │ col1 ┆ col2         ┆ col3         ┆ col4 │
    │ ---  ┆ ---          ┆ ---          ┆ ---  │
    │ i64  ┆ decimal[*,0] ┆ decimal[*,0] ┆ str  │
    ╞══════╪══════════════╪══════════════╪══════╡
    │ 1    ┆ 1            ┆ 1            ┆ a    │
    │ 2    ┆ 2            ┆ 2            ┆ b    │
    │ 3    ┆ 3            ┆ 3            ┆ c    │
    │ 4    ┆ 4            ┆ 4            ┆ d    │
    │ 5    ┆ 5            ┆ 5            ┆ e    │
    └──────┴──────────────┴──────────────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────────────┬──────┐
    │ col1 ┆ col2 ┆ col3         ┆ col4 │
    │ ---  ┆ ---  ┆ ---          ┆ ---  │
    │ i64  ┆ f32  ┆ decimal[*,0] ┆ str  │
    ╞══════╪══════╪══════════════╪══════╡
    │ 1    ┆ 1.0  ┆ 1            ┆ a    │
    │ 2    ┆ 2.0  ┆ 2            ┆ b    │
    │ 3    ┆ 3.0  ┆ 3            ┆ c    │
    │ 4    ┆ 4.0  ┆ 4            ┆ d    │
    │ 5    ┆ 5.0  ┆ 5            ┆ e    │
    └──────┴──────┴──────────────┴──────┘

    ```
    """

    def _transform_frame(self, frame: pl.DataFrame, columns: Sequence[str]) -> pl.DataFrame:
        return frame.with_columns(
            frame.select((cs.by_name(columns) & cs.decimal()).cast(self._dtype, **self._kwargs))
        )


class FloatCastTransformer(CastTransformer):
    r"""Implement a transformer to convert columns of type float to a new
    data type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        dtype: The target data type.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``cast``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import FloatCast
    >>> transformer = FloatCast(columns=["col1", "col2"], dtype=pl.Int32)
    >>> transformer
    FloatCastTransformer(columns=('col1', 'col2'), dtype=Int32, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...         "col3": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     },
    ...     schema={
    ...         "col1": pl.Int64,
    ...         "col2": pl.Float64,
    ...         "col3": pl.Float64,
    ...         "col4": pl.String,
    ...     },
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ f64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1.0  ┆ 1.0  ┆ a    │
    │ 2    ┆ 2.0  ┆ 2.0  ┆ b    │
    │ 3    ┆ 3.0  ┆ 3.0  ┆ c    │
    │ 4    ┆ 4.0  ┆ 4.0  ┆ d    │
    │ 5    ┆ 5.0  ┆ 5.0  ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i32  ┆ f64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1.0  ┆ a    │
    │ 2    ┆ 2    ┆ 2.0  ┆ b    │
    │ 3    ┆ 3    ┆ 3.0  ┆ c    │
    │ 4    ┆ 4    ┆ 4.0  ┆ d    │
    │ 5    ┆ 5    ┆ 5.0  ┆ e    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def _transform_frame(self, frame: pl.DataFrame, columns: Sequence[str]) -> pl.DataFrame:
        return frame.with_columns(
            frame.select((cs.by_name(columns) & cs.float()).cast(self._dtype, **self._kwargs))
        )


class IntegerCastTransformer(CastTransformer):
    r"""Implement a transformer to convert columns of type integer to a
    new data type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        dtype: The target data type.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``cast``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import IntegerCast
    >>> transformer = IntegerCast(columns=["col1", "col2"], dtype=pl.Float32)
    >>> transformer
    IntegerCastTransformer(columns=('col1', 'col2'), dtype=Float32, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...         "col3": [1, 2, 3, 4, 5],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     },
    ...     schema={
    ...         "col1": pl.Int64,
    ...         "col2": pl.Float64,
    ...         "col3": pl.Int64,
    ...         "col4": pl.String,
    ...     },
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1.0  ┆ 1    ┆ a    │
    │ 2    ┆ 2.0  ┆ 2    ┆ b    │
    │ 3    ┆ 3.0  ┆ 3    ┆ c    │
    │ 4    ┆ 4.0  ┆ 4    ┆ d    │
    │ 5    ┆ 5.0  ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ f64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1.0  ┆ 1.0  ┆ 1    ┆ a    │
    │ 2.0  ┆ 2.0  ┆ 2    ┆ b    │
    │ 3.0  ┆ 3.0  ┆ 3    ┆ c    │
    │ 4.0  ┆ 4.0  ┆ 4    ┆ d    │
    │ 5.0  ┆ 5.0  ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def _transform_frame(self, frame: pl.DataFrame, columns: Sequence[str]) -> pl.DataFrame:
        return frame.with_columns(
            frame.select((cs.by_name(columns) & cs.integer()).cast(self._dtype, **self._kwargs))
        )


class CategoricalCastTransformer(BaseIn1Out1Transformer):
    r"""Implement a transformer to convert a column to categorical data
    type.

    Args:
        in_col: The input column name to cast.
        out_col: The output column name.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: Additional arguments passed to
            ``polars.Categorical``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import CategoricalCast
    >>> transformer = CategoricalCast(in_col="col1", out_col="out")
    >>> transformer
    CategoricalCastTransformer(in_col='col1', out_col='out', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["a", "b", "c", "d", "e"],
    ...         "col2": [1.0, 2.0, 3.0, 4.0, 5.0],
    ...     },
    ...     schema={"col1": pl.String, "col2": pl.Float64},
    ... )
    >>> frame
    shape: (5, 2)
    ┌──────┬──────┐
    │ col1 ┆ col2 │
    │ ---  ┆ ---  │
    │ str  ┆ f64  │
    ╞══════╪══════╡
    │ a    ┆ 1.0  │
    │ b    ┆ 2.0  │
    │ c    ┆ 3.0  │
    │ d    ┆ 4.0  │
    │ e    ┆ 5.0  │
    └──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌──────┬──────┬─────┐
    │ col1 ┆ col2 ┆ out │
    │ ---  ┆ ---  ┆ --- │
    │ str  ┆ f64  ┆ cat │
    ╞══════╪══════╪═════╡
    │ a    ┆ 1.0  ┆ a   │
    │ b    ┆ 2.0  ┆ b   │
    │ c    ┆ 3.0  ┆ c   │
    │ d    ┆ 4.0  ┆ d   │
    │ e    ┆ 5.0  ┆ e   │
    └──────┴──────┴─────┘

    ```
    """

    def __init__(
        self,
        in_col: str,
        out_col: str,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            in_col=in_col,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    def get_args(self) -> dict:
        return {
            "in_col": self._in_col,
            "out_col": self._out_col,
            "exist_policy": self._exist_policy,
            "missing_policy": self._missing_policy,
        } | self._kwargs

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Casting column {self._in_col!r} to categorical column {self._out_col!r} ...")
        return frame.with_columns(
            pl.col(self._in_col).cast(pl.Categorical(**self._kwargs)).alias(self._out_col)
        )



r"""Contain transformers to filter based on the cardinality (i.e. number
of unique values) in each column."""

from __future__ import annotations

__all__ = ["FilterCardinalityTransformer"]

import logging
from typing import TYPE_CHECKING

import polars as pl
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.format import str_shape_diff

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class FilterCardinalityTransformer(BaseInNTransformer):
    r"""Implement a transformer to filter based on the cardinality (i.e.
    number of unique values) in each column.

    Args:
        columns: The columns to use to filter based on the number of
            unique values. If ``None``, it processes all the columns
            of type string.
        n_min: The minimal cardinality (included).
        n_max: The maximal cardinality (excluded).
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import FilterCardinality
    >>> transformer = FilterCardinality(columns=["col1", "col2", "col3"], n_min=2, n_max=5)
    >>> transformer
    FilterCardinalityTransformer(columns=('col1', 'col2', 'col3'), n_min=2, n_max=5, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [1, 1, 1, 1, 1],
    ...         "col3": ["a", "b", "c", "a", "b"],
    ...         "col4": [1.2, float("nan"), 3.2, None, 5.2],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  ┆ f64  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ a    ┆ 1.2  │
    │ 2    ┆ 1    ┆ b    ┆ NaN  │
    │ 3    ┆ 1    ┆ c    ┆ 3.2  │
    │ 4    ┆ 1    ┆ a    ┆ null │
    │ 5    ┆ 1    ┆ b    ┆ 5.2  │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌──────┬──────┐
    │ col3 ┆ col4 │
    │ ---  ┆ ---  │
    │ str  ┆ f64  │
    ╞══════╪══════╡
    │ a    ┆ 1.2  │
    │ b    ┆ NaN  │
    │ c    ┆ 3.2  │
    │ a    ┆ null │
    │ b    ┆ 5.2  │
    └──────┴──────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        n_min: int = 0,
        n_max: int = float("inf"),
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._n_min = n_min
        self._n_max = n_max

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    def get_args(self) -> dict:
        return {
            "columns": self._columns,
            "n_min": self._n_min,
            "n_max": self._n_max,
            "exclude_columns": self._exclude_columns,
            "missing_policy": self._missing_policy,
        }

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Filtering {len(self.find_columns(frame)):,} columns based on their "
            f"cardinality [{self._n_min}, {self._n_max})..."
        )
        columns = self.find_common_columns(frame)
        valid = frame.select(
            pl.n_unique(*columns).is_between(self._n_min, self._n_max, closed="left")
        )
        cols_to_drop = [col.name for col in valid.iter_columns() if not col[0]]
        logger.info(f"Dropping {len(cols_to_drop):,} columns: {cols_to_drop}")
        out = frame.drop(cols_to_drop)
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return out



r"""Contain utility functions to manage paths."""

from __future__ import annotations

__all__ = ["find_files", "find_parquet_files", "human_file_size", "sanitize_path"]

from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import unquote, urlparse

from grizz.utils.format import human_byte

if TYPE_CHECKING:
    from collections.abc import Callable


def human_file_size(path: Path | str, decimal: int = 2) -> str:
    r"""Get a human-readable representation of a file size.

    Args:
        path: The path to the file.
        decimal: The number of decimal digits.

    Returns:
        The file size in a human-readable format.

    Example usage:

    ```pycon

    >>> from grizz.utils.path import human_file_size
    >>> human_file_size("README.md")
    '...B'

    ```
    """
    return human_byte(size=sanitize_path(path).stat().st_size, decimal=decimal)


def sanitize_path(path: Path | str) -> Path:
    r"""Sanitize a given path.

    Args:
        path: The path to sanitize.

    Returns:
        The sanitized path.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from grizz.utils.path import sanitize_path
    >>> sanitize_path("something")
    PosixPath('.../something')
    >>> sanitize_path("")
    PosixPath('...')
    >>> sanitize_path(Path("something"))
    PosixPath('.../something')
    >>> sanitize_path(Path("something/./../"))
    PosixPath('...')

    ```
    """
    if isinstance(path, str):
        # Use urlparse to parse file URI: https://stackoverflow.com/a/15048213
        path = Path(unquote(urlparse(path).path))
    return path.expanduser().resolve()


def find_files(
    path: Path | str, filter_fn: Callable[[Path], bool], recursive: bool = True
) -> list[Path]:
    r"""Find the path of all the tar files in a given path.

    This function does not check if a path is a symbolic link so be
    careful if you are using a path with symbolic links.

    Args:
        path: The path where to look for the parquet files.
        filter_fn: The path filtering function. The function
            should return ``True`` for the path to find, and
            ``False`` otherwise.
        recursive: Indicate if it should also check the sub-folders.

    Returns:
        The tuple of path of parquet files.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from grizz.utils.path import find_files
    >>> find_files(Path("something"), filter_fn=lambda path: path.name.endswith(".txt"))
    [...]

    ```
    """
    path = sanitize_path(path)
    paths = [path] if path.is_file() else path.glob("**/*" if recursive else "*")
    return list(filter(filter_fn, [p for p in paths if p.is_file()]))


def find_parquet_files(path: Path | str, recursive: bool = True) -> list[Path]:
    r"""Find the path of all the parquet files in a given path.

    This function does not check if a path is a symbolic link so be
    careful if you are using a path with symbolic links.

    Args:
        path: The path where to look for the parquet files.
        recursive: Specifies if it should also check the sub-folders.

    Returns:
        The list of parquet files.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from grizz.utils.path import find_parquet_files
    >>> find_parquet_files(Path("something"))
    [...]

    ```
    """
    return find_files(
        path=path, filter_fn=lambda path: path.name.endswith(".parquet"), recursive=recursive
    )



r"""Contain ``polars.DataFrame`` transformers to process columns with
time values."""

from __future__ import annotations

__all__ = ["TimeToSecondTransformer", "ToTimeTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseIn1Out1Transformer, BaseInNTransformer
from grizz.utils.format import str_kwargs

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class TimeToSecondTransformer(BaseIn1Out1Transformer):
    r"""Implement a transformer to convert a column with time values to
    seconds.

    Args:
        in_col: The input column with the time value to convert.
        out_col: The output column with the time in seconds.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import datetime
    >>> import polars as pl
    >>> from grizz.transformer import TimeToSecond
    >>> transformer = TimeToSecond(in_col="time", out_col="second")
    >>> transformer
    TimeToSecondTransformer(in_col='time', out_col='second', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "time": [
    ...             datetime.time(0, 0, 1, 890000),
    ...             datetime.time(0, 1, 1, 890000),
    ...             datetime.time(1, 1, 1, 890000),
    ...             datetime.time(0, 19, 19, 890000),
    ...             datetime.time(19, 19, 19, 890000),
    ...         ],
    ...         "col": ["a", "b", "c", "d", "e"],
    ...     },
    ...     schema={"time": pl.Time, "col": pl.String},
    ... )
    >>> frame
    shape: (5, 2)
    ┌──────────────┬─────┐
    │ time         ┆ col │
    │ ---          ┆ --- │
    │ time         ┆ str │
    ╞══════════════╪═════╡
    │ 00:00:01.890 ┆ a   │
    │ 00:01:01.890 ┆ b   │
    │ 01:01:01.890 ┆ c   │
    │ 00:19:19.890 ┆ d   │
    │ 19:19:19.890 ┆ e   │
    └──────────────┴─────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌──────────────┬─────┬──────────┐
    │ time         ┆ col ┆ second   │
    │ ---          ┆ --- ┆ ---      │
    │ time         ┆ str ┆ f64      │
    ╞══════════════╪═════╪══════════╡
    │ 00:00:01.890 ┆ a   ┆ 1.89     │
    │ 00:01:01.890 ┆ b   ┆ 61.89    │
    │ 01:01:01.890 ┆ c   ┆ 3661.89  │
    │ 00:19:19.890 ┆ d   ┆ 1159.89  │
    │ 19:19:19.890 ┆ e   ┆ 69559.89 │
    └──────────────┴─────┴──────────┘

    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Converting time column {self._in_col!r} to seconds {self._out_col!r} ...")
        return frame.with_columns(
            frame.select(
                pl.col(self._in_col)
                .cast(pl.Duration)
                .dt.total_microseconds()
                .truediv(1e6)
                .alias(self._out_col)
            )
        )


class ToTimeTransformer(BaseInNTransformer):
    r"""Implement a transformer to convert some columns to a
    ``polars.Time`` type.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        format: Format to use for conversion. Refer to the
            [chrono crate documentation](https://docs.rs/chrono/latest/chrono/format/strftime/index.html)
            for the full specification. Example: ``"%H:%M:%S"``.
            If set to ``None`` (default), the format is inferred from
            the data.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``to_time``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import ToTime
    >>> transformer = ToTime(columns=["col1"], format="%H:%M:%S")
    >>> transformer
    ToTimeTransformer(columns=('col1',), format='%H:%M:%S', exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["01:01:01", "02:02:02", "12:00:01", "18:18:18", "23:59:59"],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["01:01:01", "02:02:02", "12:00:01", "18:18:18", "23:59:59"],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌──────────┬──────┬──────────┐
    │ col1     ┆ col2 ┆ col3     │
    │ ---      ┆ ---  ┆ ---      │
    │ str      ┆ str  ┆ str      │
    ╞══════════╪══════╪══════════╡
    │ 01:01:01 ┆ 1    ┆ 01:01:01 │
    │ 02:02:02 ┆ 2    ┆ 02:02:02 │
    │ 12:00:01 ┆ 3    ┆ 12:00:01 │
    │ 18:18:18 ┆ 4    ┆ 18:18:18 │
    │ 23:59:59 ┆ 5    ┆ 23:59:59 │
    └──────────┴──────┴──────────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 3)
    ┌──────────┬──────┬──────────┐
    │ col1     ┆ col2 ┆ col3     │
    │ ---      ┆ ---  ┆ ---      │
    │ time     ┆ str  ┆ str      │
    ╞══════════╪══════╪══════════╡
    │ 01:01:01 ┆ 1    ┆ 01:01:01 │
    │ 02:02:02 ┆ 2    ┆ 02:02:02 │
    │ 12:00:01 ┆ 3    ┆ 12:00:01 │
    │ 18:18:18 ┆ 4    ┆ 18:18:18 │
    │ 23:59:59 ┆ 5    ┆ 23:59:59 │
    └──────────┴──────┴──────────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        format: str | None = None,  # noqa: A002
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns, exclude_columns=exclude_columns, missing_policy=missing_policy
        )
        self._format = format
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "format": self._format,
                "exclude_columns": self._exclude_columns,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Converting {len(self.find_columns(frame)):,} columns to time | "
            f"format={self._format!r} ..."
        )
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select(
                (cs.by_name(columns) & cs.string()).str.to_time(self._format, **self._kwargs)
            )
        )



r"""Contain a function to instantiate an object from its
configuration."""

from __future__ import annotations

__all__ = ["setup_object"]

import logging
from typing import TypeVar

from objectory import factory

logger = logging.getLogger(__name__)

T = TypeVar("T")


def setup_object(obj_or_config: T | dict) -> T:
    r"""Set up an object from its configuration.

    Args:
        obj_or_config: The object or its configuration.

    Returns:
        The instantiated object.

    Example usage:

    ```pycon

    >>> from grizz.utils.factory import setup_object
    >>> obj = setup_object({"_target_": "collections.deque", "iterable": [1, 2, 1, 3]})
    >>> obj
    deque([1, 2, 1, 3])
    >>> setup_object(obj)  # Do nothing because the object is already instantiated
    deque([1, 2, 1, 3])

    ```
    """
    if isinstance(obj_or_config, dict):
        logger.info("Initializing an object from its configuration... ")
        obj_or_config = factory(**obj_or_config)
    return obj_or_config



r"""Contain ``polars.DataFrame`` transformers."""

from __future__ import annotations

__all__ = [
    "AbsDiffHorizontal",
    "AbsDiffHorizontalTransformer",
    "BaseIn1Out1Transformer",
    "BaseIn2Out1Transformer",
    "BaseInNOut1Transformer",
    "BaseInNTransformer",
    "BaseTransformer",
    "Binarizer",
    "BinarizerTransformer",
    "Cast",
    "CastTransformer",
    "CategoricalCast",
    "CategoricalCastTransformer",
    "ColumnClose",
    "ColumnCloseTransformer",
    "ColumnEqual",
    "ColumnEqualMissing",
    "ColumnEqualMissingTransformer",
    "ColumnEqualTransformer",
    "ColumnGreater",
    "ColumnGreaterEqual",
    "ColumnGreaterEqualTransformer",
    "ColumnGreaterTransformer",
    "ColumnLower",
    "ColumnLowerEqual",
    "ColumnLowerEqualTransformer",
    "ColumnLowerTransformer",
    "ColumnNotEqual",
    "ColumnNotEqualMissing",
    "ColumnNotEqualMissingTransformer",
    "ColumnNotEqualTransformer",
    "ColumnSelection",
    "ColumnSelectionTransformer",
    "ConcatColumns",
    "ConcatColumnsTransformer",
    "CopyColumn",
    "CopyColumnTransformer",
    "CopyColumns",
    "CopyColumnsTransformer",
    "DecimalCast",
    "DecimalCastTransformer",
    "Diff",
    "DiffHorizontal",
    "DiffHorizontalTransformer",
    "DiffTransformer",
    "DropDuplicate",
    "DropDuplicateTransformer",
    "DropNullColumn",
    "DropNullColumnTransformer",
    "DropNullRow",
    "DropNullRowTransformer",
    "Equal",
    "EqualMissing",
    "EqualMissingTransformer",
    "EqualTransformer",
    "FillNan",
    "FillNanTransformer",
    "FillNull",
    "FillNullTransformer",
    "FilterCardinality",
    "FilterCardinalityTransformer",
    "FirstRow",
    "FirstRowTransformer",
    "FloatCast",
    "FloatCastTransformer",
    "Function",
    "FunctionTransformer",
    "Greater",
    "GreaterEqual",
    "GreaterEqualTransformer",
    "GreaterTransformer",
    "IntegerCast",
    "IntegerCastTransformer",
    "JsonDecode",
    "JsonDecodeTransformer",
    "LabelEncoder",
    "LabelEncoderTransformer",
    "Lower",
    "LowerEqual",
    "LowerEqualTransformer",
    "LowerTransformer",
    "MaxAbsScaler",
    "MaxAbsScalerTransformer",
    "MaxHorizontal",
    "MaxHorizontalTransformer",
    "MeanHorizontal",
    "MeanHorizontalTransformer",
    "MinHorizontal",
    "MinHorizontalTransformer",
    "MinMaxScaler",
    "MinMaxScalerTransformer",
    "Normalizer",
    "NormalizerTransformer",
    "NotEqual",
    "NotEqualMissing",
    "NotEqualMissingTransformer",
    "NotEqualTransformer",
    "OrdinalEncoder",
    "OrdinalEncoderTransformer",
    "PowerTransformer",
    "QuantileTransformer",
    "Replace",
    "ReplaceStrict",
    "ReplaceStrictTransformer",
    "ReplaceTransformer",
    "RobustScaler",
    "RobustScalerTransformer",
    "Sequential",
    "SequentialTransformer",
    "ShrinkMemory",
    "ShrinkMemoryTransformer",
    "SimpleImputer",
    "SimpleImputerTransformer",
    "Sort",
    "SortColumns",
    "SortColumnsTransformer",
    "SortTransformer",
    "SqlTransformer",
    "StandardScaler",
    "StandardScalerTransformer",
    "StripChars",
    "StripCharsTransformer",
    "SumHorizontal",
    "SumHorizontalTransformer",
    "TimeDiff",
    "TimeDiffTransformer",
    "TimeToSecond",
    "TimeToSecondTransformer",
    "ToDatetime",
    "ToDatetimeTransformer",
    "ToTime",
    "ToTimeTransformer",
    "is_transformer_config",
    "setup_transformer",
]

from grizz.transformer.base import (
    BaseTransformer,
    is_transformer_config,
    setup_transformer,
)
from grizz.transformer.binarizer import BinarizerTransformer
from grizz.transformer.binarizer import BinarizerTransformer as Binarizer
from grizz.transformer.cardinality import FilterCardinalityTransformer
from grizz.transformer.cardinality import (
    FilterCardinalityTransformer as FilterCardinality,
)
from grizz.transformer.casting import CastTransformer
from grizz.transformer.casting import CastTransformer as Cast
from grizz.transformer.casting import CategoricalCastTransformer
from grizz.transformer.casting import CategoricalCastTransformer as CategoricalCast
from grizz.transformer.casting import DecimalCastTransformer
from grizz.transformer.casting import DecimalCastTransformer as DecimalCast
from grizz.transformer.casting import FloatCastTransformer
from grizz.transformer.casting import FloatCastTransformer as FloatCast
from grizz.transformer.casting import IntegerCastTransformer
from grizz.transformer.casting import IntegerCastTransformer as IntegerCast
from grizz.transformer.close import ColumnCloseTransformer
from grizz.transformer.close import ColumnCloseTransformer as ColumnClose
from grizz.transformer.column_comparison import ColumnEqualMissingTransformer
from grizz.transformer.column_comparison import (
    ColumnEqualMissingTransformer as ColumnEqualMissing,
)
from grizz.transformer.column_comparison import ColumnEqualTransformer
from grizz.transformer.column_comparison import ColumnEqualTransformer as ColumnEqual
from grizz.transformer.column_comparison import ColumnGreaterEqualTransformer
from grizz.transformer.column_comparison import (
    ColumnGreaterEqualTransformer as ColumnGreaterEqual,
)
from grizz.transformer.column_comparison import ColumnGreaterTransformer
from grizz.transformer.column_comparison import (
    ColumnGreaterTransformer as ColumnGreater,
)
from grizz.transformer.column_comparison import ColumnLowerEqualTransformer
from grizz.transformer.column_comparison import (
    ColumnLowerEqualTransformer as ColumnLowerEqual,
)
from grizz.transformer.column_comparison import ColumnLowerTransformer
from grizz.transformer.column_comparison import ColumnLowerTransformer as ColumnLower
from grizz.transformer.column_comparison import ColumnNotEqualMissingTransformer
from grizz.transformer.column_comparison import (
    ColumnNotEqualMissingTransformer as ColumnNotEqualMissing,
)
from grizz.transformer.column_comparison import ColumnNotEqualTransformer
from grizz.transformer.column_comparison import (
    ColumnNotEqualTransformer as ColumnNotEqual,
)
from grizz.transformer.columns import (
    BaseIn1Out1Transformer,
    BaseIn2Out1Transformer,
    BaseInNOut1Transformer,
    BaseInNTransformer,
)
from grizz.transformer.comparison import EqualMissingTransformer
from grizz.transformer.comparison import EqualMissingTransformer as EqualMissing
from grizz.transformer.comparison import EqualTransformer
from grizz.transformer.comparison import EqualTransformer as Equal
from grizz.transformer.comparison import GreaterEqualTransformer
from grizz.transformer.comparison import GreaterEqualTransformer as GreaterEqual
from grizz.transformer.comparison import GreaterTransformer
from grizz.transformer.comparison import GreaterTransformer as Greater
from grizz.transformer.comparison import LowerEqualTransformer
from grizz.transformer.comparison import LowerEqualTransformer as LowerEqual
from grizz.transformer.comparison import LowerTransformer
from grizz.transformer.comparison import LowerTransformer as Lower
from grizz.transformer.comparison import NotEqualMissingTransformer
from grizz.transformer.comparison import NotEqualMissingTransformer as NotEqualMissing
from grizz.transformer.comparison import NotEqualTransformer
from grizz.transformer.comparison import NotEqualTransformer as NotEqual
from grizz.transformer.concat import ConcatColumnsTransformer
from grizz.transformer.concat import ConcatColumnsTransformer as ConcatColumns
from grizz.transformer.copy import CopyColumnsTransformer
from grizz.transformer.copy import CopyColumnsTransformer as CopyColumns
from grizz.transformer.copy import CopyColumnTransformer
from grizz.transformer.copy import CopyColumnTransformer as CopyColumn
from grizz.transformer.datetime import ToDatetimeTransformer
from grizz.transformer.datetime import ToDatetimeTransformer as ToDatetime
from grizz.transformer.diff import DiffTransformer
from grizz.transformer.diff import DiffTransformer as Diff
from grizz.transformer.diff import TimeDiffTransformer
from grizz.transformer.diff import TimeDiffTransformer as TimeDiff
from grizz.transformer.diff_horizontal import AbsDiffHorizontalTransformer
from grizz.transformer.diff_horizontal import (
    AbsDiffHorizontalTransformer as AbsDiffHorizontal,
)
from grizz.transformer.diff_horizontal import DiffHorizontalTransformer
from grizz.transformer.diff_horizontal import (
    DiffHorizontalTransformer as DiffHorizontal,
)
from grizz.transformer.duplicate import DropDuplicateTransformer
from grizz.transformer.duplicate import DropDuplicateTransformer as DropDuplicate
from grizz.transformer.fill import FillNanTransformer
from grizz.transformer.fill import FillNanTransformer as FillNan
from grizz.transformer.fill import FillNullTransformer
from grizz.transformer.fill import FillNullTransformer as FillNull
from grizz.transformer.function import FunctionTransformer
from grizz.transformer.function import FunctionTransformer as Function
from grizz.transformer.impute import SimpleImputerTransformer
from grizz.transformer.impute import SimpleImputerTransformer as SimpleImputer
from grizz.transformer.json import JsonDecodeTransformer
from grizz.transformer.json import JsonDecodeTransformer as JsonDecode
from grizz.transformer.label_encoder import LabelEncoderTransformer
from grizz.transformer.label_encoder import LabelEncoderTransformer as LabelEncoder
from grizz.transformer.max import MaxHorizontalTransformer
from grizz.transformer.max import MaxHorizontalTransformer as MaxHorizontal
from grizz.transformer.max_scaler import MaxAbsScalerTransformer
from grizz.transformer.max_scaler import MaxAbsScalerTransformer as MaxAbsScaler
from grizz.transformer.mean import MeanHorizontalTransformer
from grizz.transformer.mean import MeanHorizontalTransformer as MeanHorizontal
from grizz.transformer.min import MinHorizontalTransformer
from grizz.transformer.min import MinHorizontalTransformer as MinHorizontal
from grizz.transformer.min_max_scaler import MinMaxScalerTransformer
from grizz.transformer.min_max_scaler import MinMaxScalerTransformer as MinMaxScaler
from grizz.transformer.normalizer import NormalizerTransformer
from grizz.transformer.normalizer import NormalizerTransformer as Normalizer
from grizz.transformer.null import DropNullColumnTransformer
from grizz.transformer.null import DropNullColumnTransformer as DropNullColumn
from grizz.transformer.null import DropNullRowTransformer
from grizz.transformer.null import DropNullRowTransformer as DropNullRow
from grizz.transformer.ordinal_encoder import OrdinalEncoderTransformer
from grizz.transformer.ordinal_encoder import (
    OrdinalEncoderTransformer as OrdinalEncoder,
)
from grizz.transformer.power import PowerTransformer
from grizz.transformer.quantile import QuantileTransformer
from grizz.transformer.replace import ReplaceStrictTransformer
from grizz.transformer.replace import ReplaceStrictTransformer as ReplaceStrict
from grizz.transformer.replace import ReplaceTransformer
from grizz.transformer.replace import ReplaceTransformer as Replace
from grizz.transformer.robust_scaler import RobustScalerTransformer
from grizz.transformer.robust_scaler import RobustScalerTransformer as RobustScaler
from grizz.transformer.row import FirstRowTransformer
from grizz.transformer.row import FirstRowTransformer as FirstRow
from grizz.transformer.selection import ColumnSelectionTransformer
from grizz.transformer.selection import ColumnSelectionTransformer as ColumnSelection
from grizz.transformer.sequential import SequentialTransformer
from grizz.transformer.sequential import SequentialTransformer as Sequential
from grizz.transformer.shrink import ShrinkMemoryTransformer
from grizz.transformer.shrink import ShrinkMemoryTransformer as ShrinkMemory
from grizz.transformer.sorting import SortColumnsTransformer
from grizz.transformer.sorting import SortColumnsTransformer as SortColumns
from grizz.transformer.sorting import SortTransformer
from grizz.transformer.sorting import SortTransformer as Sort
from grizz.transformer.sql import SqlTransformer
from grizz.transformer.standard_scaler import StandardScalerTransformer
from grizz.transformer.standard_scaler import (
    StandardScalerTransformer as StandardScaler,
)
from grizz.transformer.string import StripCharsTransformer
from grizz.transformer.string import StripCharsTransformer as StripChars
from grizz.transformer.sum import SumHorizontalTransformer
from grizz.transformer.sum import SumHorizontalTransformer as SumHorizontal
from grizz.transformer.time import TimeToSecondTransformer
from grizz.transformer.time import TimeToSecondTransformer as TimeToSecond
from grizz.transformer.time import ToTimeTransformer
from grizz.transformer.time import ToTimeTransformer as ToTime



r"""Contain utility functions to format strings."""

from __future__ import annotations

__all__ = [
    "human_byte",
    "str_boolean_series_stats",
    "str_col_diff",
    "str_kwargs",
    "str_row_diff",
    "str_shape_diff",
]

from typing import TYPE_CHECKING

from grizz.utils.series import compute_stats_boolean

if TYPE_CHECKING:
    from collections.abc import Mapping

    import polars as pl


def human_byte(size: float, decimal: int = 2) -> str:
    r"""Return a human-readable string representation of byte sizes.

    Args:
        size: The number of bytes.
        decimal: The number of decimal digits.

    Returns:
        The human-readable string representation of byte sizes.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import human_byte
    >>> human_byte(2)
    '2.00 B'
    >>> human_byte(2048)
    '2.00 KB'
    >>> human_byte(2097152)
    '2.00 MB'

    ```
    """
    for unit in ["B", "KB", "MB", "GB", "TB", "PB"]:
        if abs(size) < 1024.0:
            break
        if unit != "PB":
            size /= 1024.0
    return f"{size:,.{decimal}f} {unit}"


def str_kwargs(mapping: Mapping) -> str:
    r"""Return a string of the input mapping.

    This function is designed to be used in ``__repr__`` and
    ``__str__`` methods.

    Args:
        mapping: The mapping.

    Returns:
        The generated string.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import str_kwargs
    >>> str_kwargs({"key1": 1})
    ', key1=1'
    >>> str_kwargs({"key1": 1, "key2": 2})
    ', key1=1, key2=2'

    ```
    """
    args = ", ".join([f"{key}={value}" for key, value in mapping.items()])
    if args:
        args = ", " + args
    return args


def str_col_diff(orig: int, final: int) -> str:
    r"""Return a string that indicates the difference of columns.

    Args:
        orig: The original number of columns.
        final: The final number of columns.

    Returns:
        The generated string with the difference of columns.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import str_col_diff
    >>> str_col_diff(100, 10)
    90/100 (90.0000 %) columns have been removed
    >>> str_col_diff(100, 99)
    1/100 (1.0000 %) column has been removed

    ```
    """
    diff = abs(orig - final)
    desc = "removed" if orig >= final else "added"
    diff_pct = 100 * diff / orig if orig > 0 else float("nan")
    row = "columns have" if diff > 1 else "column has"
    return f"{diff:,}/{orig:,} ({diff_pct:.4f} %) {row} been {desc}"


def str_row_diff(orig: int, final: int) -> str:
    r"""Return a string that indicates the difference of rows.

    Args:
        orig: The original number of rows.
        final: The final number of rows.

    Returns:
        The generated string with the difference of rows.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import str_row_diff
    >>> str_row_diff(100, 10)
    90/100 (90.0000 %) rows have been removed
    >>> str_row_diff(100, 99)
    1/100 (1.0000 %) row has been removed

    ```
    """
    diff = abs(orig - final)
    desc = "removed" if orig >= final else "added"
    diff_pct = 100 * diff / orig if orig > 0 else float("nan")
    row = "rows have" if diff > 1 else "row has"
    return f"{diff:,}/{orig:,} ({diff_pct:.4f} %) {row} been {desc}"


def str_shape_diff(orig: tuple[int, int], final: tuple[int, int]) -> str:
    r"""Return a string that indicates the difference of DataFrame
    shapes.

    Args:
        orig: The original shape.
        final: The final shape.

    Returns:
        The generated string with the difference of DataFrame shapes.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import str_shape_diff
    >>> str_shape_diff(orig=(100, 10), final=(80, 8))
    DataFrame shape: (100, 10) -> (80, 8) | 20/100 (20.0000 %) rows have been removed |
    2/10 (20.0000 %) columns have been removed

    ```
    """
    msg = f"DataFrame shape: {orig} -> {final}"
    if orig[0] != final[0]:
        msg += f" | {str_row_diff(orig=orig[0], final=final[0])}"
    if orig[1] != final[1]:
        msg += f" | {str_col_diff(orig=orig[1], final=final[1])}"
    return msg


def str_size_diff(orig: float, final: float) -> str:
    r"""Return a string that indicates the difference of DataFrame sizes.

    Args:
        orig: The original size.
        final: The final size.

    Returns:
        The generated string with the difference of DataFrame sizes.

    Example usage:

    ```pycon

    >>> from grizz.utils.format import str_size_diff
    >>> str_size_diff(orig=100, final=120)
    DataFrame estimated size: 100.00 B -> 120.00 B | difference: 20.00 B (20.0000 %)

    ```
    """
    diff = final - orig
    pct = 100 * diff / orig if diff != 0 else float("nan")
    return (
        f"DataFrame estimated size: {human_byte(orig)} -> {human_byte(final)} | "
        f"difference: {human_byte(diff)} ({pct:.4f} %)"
    )


def str_boolean_series_stats(series: pl.Series) -> str:
    r"""Return a string about the content of a Boolean series.

    Args:
        series: The input series.

    Returns:
        The generated string about the input series.

    Raises:
        ValueError: if ``series`` is not a Boolean series.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.format import str_boolean_series_stats
    >>> series = pl.Series([True, False, None, None, False, None])
    >>> str_boolean_series_stats(series)
    true: 1/3 (33.3333 %) | null: 3/6 (50.0000 %)

    ```
    """
    stats = compute_stats_boolean(series)
    total = stats["total"]
    num_null = stats["num_null"]
    num_non_null = stats["total"] - num_null
    pct_true = 100 * stats["num_true"] / num_non_null if num_non_null > 0 else float("nan")
    pct_null = 100 * num_null / total if total > 0 else float("nan")
    return (
        f"true: {stats['num_true']:,}/{num_non_null:,} ({pct_true:.4f} %) | "
        f"null: {num_null:,}/{total:,} ({pct_null:.4f} %)"
    )



r"""Contain ``polars.DataFrame`` transformers to sort the DataFrame."""

from __future__ import annotations

__all__ = ["SortColumnsTransformer", "SortTransformer"]

import logging
from typing import TYPE_CHECKING, Any

from coola.utils.format import repr_mapping_line

from grizz.transformer.base import BaseTransformer
from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.format import str_kwargs

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl


class SortTransformer(BaseInNTransformer):
    r"""Implement a transformer to sort the DataFrame by the given
    columns.

    Args:
        columns: The columns to use to sort the rows.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments to pass to ``sort``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Sort
    >>> transformer = Sort(columns=["col3", "col1"])
    >>> transformer
    SortTransformer(columns=('col3', 'col1'), exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {"col1": [1, 2, None], "col2": [6.0, 5.0, 4.0], "col3": ["a", "c", "b"]}
    ... )
    >>> frame
    shape: (3, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 6.0  ┆ a    │
    │ 2    ┆ 5.0  ┆ c    │
    │ null ┆ 4.0  ┆ b    │
    └──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (3, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 6.0  ┆ a    │
    │ null ┆ 4.0  ┆ b    │
    │ 2    ┆ 5.0  ┆ c    │
    └──────┴──────┴──────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "exclude_columns": self._exclude_columns,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        cols = self.find_columns(frame)
        logger.info(f"Sorting rows based on {len(cols):,} columns: {cols}")
        # Note: it is not possible to use find_common_columns because find_common_columns
        # may change the order of the columns.
        columns = self._find_existing_columns(frame)
        return frame.sort(columns, **self._kwargs)

    def _find_existing_columns(self, frame: pl.DataFrame) -> list[str]:
        cols = self.find_columns(frame)
        return [col for col in cols if col in frame]


class SortColumnsTransformer(BaseTransformer):
    r"""Implement a transformer to sort the DataFrame columns by name.

    Args:
        reverse: If set to ``False``, then the columns are sorted by
            alphabetical order.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import SortColumns
    >>> transformer = SortColumns()
    >>> transformer
    SortColumnsTransformer(reverse=False)
    >>> frame = pl.DataFrame(
    ...     {"col2": [1, 2, None], "col3": [6.0, 5.0, 4.0], "col1": ["a", "c", "b"]}
    ... )
    >>> frame
    shape: (3, 3)
    ┌──────┬──────┬──────┐
    │ col2 ┆ col3 ┆ col1 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 6.0  ┆ a    │
    │ 2    ┆ 5.0  ┆ c    │
    │ null ┆ 4.0  ┆ b    │
    └──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (3, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ str  ┆ i64  ┆ f64  │
    ╞══════╪══════╪══════╡
    │ a    ┆ 1    ┆ 6.0  │
    │ c    ┆ 2    ┆ 5.0  │
    │ b    ┆ null ┆ 4.0  │
    └──────┴──────┴──────┘

    ```
    """

    def __init__(self, reverse: bool = False) -> None:
        self._reverse = reverse

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(reverse={self._reverse})"

    def fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def fit_transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self.fit(frame)
        return self.transform(frame)

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Sorting columns | reverse={self._reverse} ...")
        return frame.select(sorted(frame.columns, reverse=self._reverse))



r"""Contain the implementation of a ClickHouse ingestor."""

from __future__ import annotations

__all__ = ["ClickHouseArrowIngestor"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
from coola import objects_are_equal
from iden.utils.time import timeblock

from grizz.ingestor.base import BaseIngestor
from grizz.utils.factory import setup_object
from grizz.utils.format import human_byte
from grizz.utils.imports import check_clickhouse_connect, check_pyarrow

if TYPE_CHECKING:
    from grizz.utils.imports import is_clickhouse_connect_available

    if is_clickhouse_connect_available():
        import clickhouse_connect


logger = logging.getLogger(__name__)


class ClickHouseArrowIngestor(BaseIngestor):
    r"""Implement a ClickHouse DataFrame ingestor that uses Arrow.

    This ingestor requires ``clickhouse_connect`` and ``pyarrow``.

    Args:
        query: The query to get the data.
        client: The ClickHouse client or its configuration.
            Please check the documentation of
            ``clickhouse_connect.get_client`` to get more information.

    Example usage:

    ```pycon

    >>> from grizz.ingestor import ClickHouseArrowIngestor
    >>> import clickhouse_connect
    >>> client = clickhouse_connect.get_client()  # doctest: +SKIP
    >>> ingestor = ClickHouseArrowIngestor(query="", client=client)  # doctest: +SKIP
    >>> frame = ingestor.ingest()  # doctest: +SKIP

    ```
    """

    def __init__(self, query: str, client: clickhouse_connect.driver.Client | dict) -> None:
        check_clickhouse_connect()
        check_pyarrow()
        self._query = str(query)
        self._client: clickhouse_connect.driver.Client = setup_object(client)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._query == other._query and objects_are_equal(
            self._client, other._client, equal_nan=equal_nan
        )

    def ingest(self) -> pl.DataFrame:
        logger.info(
            f"Ingesting data from ClickHouse... \n\n"
            "---------------------------------------------------------------------------------\n"
            f"query:\n{self._query}\n"
            "---------------------------------------------------------------------------------\n\n"
        )
        with timeblock("DataFrame ingestion time: {time}"):
            data = self._client.query_arrow(query=self._query)
            frame = pl.from_arrow(data)
            frame = frame.select(sorted(frame.columns))
            logger.info(
                f"DataFrame ingested | shape={frame.shape}  "
                f"estimated size={human_byte(frame.estimated_size())}"
            )
        logger.info(f"number of unique column names: {len(set(frame.columns)):,}")
        return frame



r"""Contain transformers to drop columns or rows with null values."""

from __future__ import annotations

__all__ = ["DropNullColumnTransformer", "DropNullRowTransformer"]

import logging
from itertools import compress
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.format import str_kwargs, str_shape_diff

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class DropNullColumnTransformer(BaseInNTransformer):
    r"""Implement a transformer to remove the columns that have too many
    null values.

    Args:
        columns: The columns to convert. ``None`` means all the
            columns.
        threshold: The maximum percentage of null values to keep
            columns. If the proportion of null vallues is greater
            or equal to this threshold value, the column is removed.
            If set to ``1.0``, it removes all the columns that have
            only null values.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``cast``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import DropNullColumn
    >>> transformer = DropNullColumn()
    >>> transformer
    DropNullColumnTransformer(columns=None, threshold=1.0, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
    ...         "col2": [1, None, 3, None, 5],
    ...         "col3": [None, None, None, None, None],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌────────────┬──────┬──────┐
    │ col1       ┆ col2 ┆ col3 │
    │ ---        ┆ ---  ┆ ---  │
    │ str        ┆ i64  ┆ null │
    ╞════════════╪══════╪══════╡
    │ 2020-1-1   ┆ 1    ┆ null │
    │ 2020-1-2   ┆ null ┆ null │
    │ 2020-1-31  ┆ 3    ┆ null │
    │ 2020-12-31 ┆ null ┆ null │
    │ null       ┆ 5    ┆ null │
    └────────────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌────────────┬──────┐
    │ col1       ┆ col2 │
    │ ---        ┆ ---  │
    │ str        ┆ i64  │
    ╞════════════╪══════╡
    │ 2020-1-1   ┆ 1    │
    │ 2020-1-2   ┆ null │
    │ 2020-1-31  ┆ 3    │
    │ 2020-12-31 ┆ null │
    │ null       ┆ 5    │
    └────────────┴──────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        threshold: float = 1.0,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns, exclude_columns=exclude_columns, missing_policy=missing_policy
        )
        self._threshold = threshold
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "threshold": self._threshold,
                "exclude_columns": self._exclude_columns,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Checking columns and dropping the columns that have too "
            f"many null values (threshold={self._threshold})..."
        )
        if frame.is_empty():
            return frame
        columns = self.find_common_columns(frame)
        pct = frame.select(columns).null_count() / frame.shape[0]
        cols = list(compress(pct.columns, (pct >= self._threshold).row(0)))
        logger.info(
            f"Dropping {len(cols):,} columns that have too "
            f"many null values (threshold={self._threshold})..."
        )
        logger.info(f"dropped columns: {cols}")
        out = frame.drop(cols)
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return out


class DropNullRowTransformer(BaseInNTransformer):
    r"""Implement a transformer to drop all rows that contain null
    values.

    Note that all the values in the row need to be null to drop the
    row.

    Args:
        columns: The columns to check. If set to ``None`` (default),
            use all columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import DropNullRow
    >>> transformer = DropNullRow()
    >>> transformer
    DropNullRowTransformer(columns=None, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
    ...         "col2": [1, None, 3, None, None],
    ...         "col3": [None, None, None, None, None],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌────────────┬──────┬──────┐
    │ col1       ┆ col2 ┆ col3 │
    │ ---        ┆ ---  ┆ ---  │
    │ str        ┆ i64  ┆ null │
    ╞════════════╪══════╪══════╡
    │ 2020-1-1   ┆ 1    ┆ null │
    │ 2020-1-2   ┆ null ┆ null │
    │ 2020-1-31  ┆ 3    ┆ null │
    │ 2020-12-31 ┆ null ┆ null │
    │ null       ┆ null ┆ null │
    └────────────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (4, 3)
    ┌────────────┬──────┬──────┐
    │ col1       ┆ col2 ┆ col3 │
    │ ---        ┆ ---  ┆ ---  │
    │ str        ┆ i64  ┆ null │
    ╞════════════╪══════╪══════╡
    │ 2020-1-1   ┆ 1    ┆ null │
    │ 2020-1-2   ┆ null ┆ null │
    │ 2020-1-31  ┆ 3    ┆ null │
    │ 2020-12-31 ┆ null ┆ null │
    └────────────┴──────┴──────┘

    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Dropping all rows that contain null values in {len(self.find_columns(frame)):,} "
            "columns...."
        )
        columns = self.find_common_columns(frame)
        out = frame.filter(~pl.all_horizontal(cs.by_name(columns).is_null()))
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return out



r"""Contain ``polars.DataFrame`` transformers to process string
values."""

from __future__ import annotations

__all__ = ["StripCharsTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.format import str_kwargs

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class StripCharsTransformer(BaseInNTransformer):
    r"""Implement a transformer to remove leading and trailing
    characters.

    Args:
        columns: The columns to prepare. If ``None``, it processes all
            the columns of type string.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``strip_chars``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import StripChars
    >>> transformer = StripChars(columns=["col2", "col3"])
    >>> transformer
    StripCharsTransformer(columns=('col2', 'col3'), exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬───────┬───────┐
    │ col1 ┆ col2 ┆ col3  ┆ col4  │
    │ ---  ┆ ---  ┆ ---   ┆ ---   │
    │ i64  ┆ str  ┆ str   ┆ str   │
    ╞══════╪══════╪═══════╪═══════╡
    │ 1    ┆ 1    ┆ a     ┆ a     │
    │ 2    ┆ 2    ┆  b    ┆  b    │
    │ 3    ┆ 3    ┆   c   ┆   c   │
    │ 4    ┆ 4    ┆ d     ┆ d     │
    │ 5    ┆ 5    ┆ e     ┆ e     │
    └──────┴──────┴───────┴───────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬───────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4  │
    │ ---  ┆ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ str  ┆ str  ┆ str   │
    ╞══════╪══════╪══════╪═══════╡
    │ 1    ┆ 1    ┆ a    ┆ a     │
    │ 2    ┆ 2    ┆ b    ┆  b    │
    │ 3    ┆ 3    ┆ c    ┆   c   │
    │ 4    ┆ 4    ┆ d    ┆ d     │
    │ 5    ┆ 5    ┆ e    ┆ e     │
    └──────┴──────┴──────┴───────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "exclude_columns": self._exclude_columns,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Stripping characters of {len(self.find_columns(frame)):,} columns...")
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select((cs.by_name(columns) & cs.string()).str.strip_chars(**self._kwargs))
        )



r"""Contain a transformer that is a wrapper around a function to
transform the DataFrame."""

from __future__ import annotations

__all__ = ["FunctionTransformer"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from grizz.transformer.base import BaseTransformer

if TYPE_CHECKING:
    from collections.abc import Callable

    import polars as pl

logger = logging.getLogger(__name__)


class FunctionTransformer(BaseTransformer):
    r"""Implement a transformer that is a wrapper around a function to
    transform the DataFrame.

    Args:
        func: The function to transform the DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import FunctionTransformer
    >>> transformer = FunctionTransformer(
    ...     func=lambda frame: frame.filter(pl.col("col1").is_in({2, 4}))
    ... )
    >>> transformer
    FunctionTransformer(func=<function <lambda> at 0x...>)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> out = transformer.transform(frame)
    >>> out
    shape: (2, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def __init__(self, func: Callable[[pl.DataFrame], pl.DataFrame]) -> None:
        self._func = func

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(func={self._func})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._func, other._func, equal_nan=equal_nan)

    def fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def fit_transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self.fit(frame)
        return self.transform(frame)

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        return self._func(frame)



r"""Contain utility functions for series."""

from __future__ import annotations

__all__ = ["compute_stats_boolean"]


import polars as pl


def compute_stats_boolean(series: pl.Series) -> dict[str, float]:
    r"""Compute some basic statistics about a Boolean series.

    Args:
        series: The series to analyze.

    Returns:
        The statistics about the input Boolean series.

    Raises:
        ValueError: if ``series`` is not a Boolean series.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.series import compute_stats_boolean
    >>> series = pl.Series([True, False, None, None, False, None])
    >>> compute_stats_boolean(series)
    {'num_false': 2, 'num_null': 3, 'num_true': 1, 'total': 6}

    ```
    """
    if series.dtype != pl.Boolean:
        msg = f"Incorrect dtype. Expected a Boolean series but received {series.dtype}"
        raise ValueError(msg)
    num_null = series.is_null().sum()
    num_true = series.sum()
    total = series.shape[0]
    return {
        "num_false": total - num_true - num_null,
        "num_null": num_null,
        "num_true": num_true,
        "total": total,
    }



r"""Contain a ``polars.DataFrame`` transformer to indicate if the values
of two columns are element-wise equal within a tolerance."""

from __future__ import annotations

__all__ = ["ColumnCloseTransformer"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseIn2Out1Transformer
from grizz.utils.format import str_boolean_series_stats

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class ColumnCloseTransformer(BaseIn2Out1Transformer):
    r"""Implement a transformer to compute a column that indicates if the
    values of two columns are element-wise equal within a tolerance.

    The output column contains ``True`` if two columns are element-wise
    equal within a tolerance. Internally, this tranformer computes:
    ``out = (|actual - expected| <= atol + rtol * |expected|)``

    Args:
        actual: The actual input column name. This column must be a
            numeric column.
        expected: The expected input column name. This column must be
            a numeric column.
        out_col: The output column name.
        atol: The absolute tolerance parameter.
        rtol: The relative tolerance parameter.
        equal_nan: Whether to compare NaN's as equal. If ``True``,
            NaN's in ``actual`` will be considered equal to NaN's in
            ``expected`` in the output column.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import ColumnClose
    >>> transformer = ColumnClose(actual="col1", expected="col2", out_col="out")
    >>> transformer
    ColumnCloseTransformer(actual='col1', expected='col2', out_col='out', atol=1e-08, rtol=1e-05, equal_nan=False, exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [5, 4, 3, 2, 1],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 5    ┆ a    │
    │ 2    ┆ 4    ┆ b    │
    │ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 2    ┆ d    │
    │ 5    ┆ 1    ┆ e    │
    └──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬───────┐
    │ col1 ┆ col2 ┆ col3 ┆ out   │
    │ ---  ┆ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ i64  ┆ str  ┆ bool  │
    ╞══════╪══════╪══════╪═══════╡
    │ 1    ┆ 5    ┆ a    ┆ false │
    │ 2    ┆ 4    ┆ b    ┆ false │
    │ 3    ┆ 3    ┆ c    ┆ true  │
    │ 4    ┆ 2    ┆ d    ┆ false │
    │ 5    ┆ 1    ┆ e    ┆ false │
    └──────┴──────┴──────┴───────┘

    ```
    """

    def __init__(
        self,
        actual: str,
        expected: str,
        out_col: str,
        atol: float = 1e-8,
        rtol: float = 1e-5,
        equal_nan: bool = False,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            in1_col=actual,
            in2_col=expected,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )
        self._atol = float(atol)
        self._rtol = float(rtol)
        self._equal_nan = equal_nan

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    def get_args(self) -> dict:
        return {
            "actual": self._in1_col,
            "expected": self._in2_col,
            "out_col": self._out_col,
            "atol": self._atol,
            "rtol": self._rtol,
            "equal_nan": self._equal_nan,
            "exist_policy": self._exist_policy,
            "missing_policy": self._missing_policy,
        }

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Computing the equality within tolerance between actual column {self._in1_col!r} "
            f"and expected column {self._in2_col!r} | out_col={self._out_col!r} | "
            f"atol={self._atol}  rtol={self._rtol}  equal_nan={self._equal_nan}"
        )
        diff = (frame[self._in1_col] - frame[self._in2_col]).abs()
        tol = frame[self._in2_col].abs() * self._rtol + self._atol
        tol_check = (diff <= tol) & ~frame[self._in2_col].is_nan()

        if self._equal_nan:
            nan_check = frame[self._in1_col].is_nan() & frame[self._in2_col].is_nan()
            tol_check = tol_check | nan_check
        logger.info(f"column: {self._out_col!r} | {str_boolean_series_stats(tol_check)}")
        return frame.with_columns(tol_check.alias(self._out_col))



r"""Contain no-op functions."""

from __future__ import annotations

__all__ = ["tqdm"]

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable


def tqdm(iterable: Iterable, *args: Any, **kwargs: Any) -> Iterable:  # noqa: ARG001
    r"""Implement a no-op tqdm progressbar that is used when tqdm is not
    installed.

    Args:
        iterable: Iterable to decorate with a progressbar.
        *args: Positional arbitrary arguments.
        **kwargs: Keyword arbitrary arguments.

    Returns:
        The input iterable.
    """
    return iterable



r"""Contain ``polars.DataFrame`` transformers to copy columns."""

from __future__ import annotations

__all__ = ["CopyColumnTransformer", "CopyColumnsTransformer"]

import logging
from typing import TYPE_CHECKING

import polars as pl

from grizz.transformer.columns import BaseIn1Out1Transformer, BaseInNTransformer
from grizz.utils.column import check_column_exist_policy, check_existing_columns
from grizz.utils.format import str_shape_diff

if TYPE_CHECKING:
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class CopyColumnTransformer(BaseIn1Out1Transformer):
    r"""Implement a ``polars.DataFrame`` to copy a column.

    Args:
        in_col: The input column name i.e. the column to copy.
        out_col: The output column name i.e. the copied column.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import CopyColumn
    >>> transformer = CopyColumn(in_col="col1", out_col="out")
    >>> transformer
    CopyColumnTransformer(in_col='col1', out_col='out', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 5)
    ┌──────┬──────┬──────┬──────┬─────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ out │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ --- │
    │ i64  ┆ str  ┆ str  ┆ str  ┆ i64 │
    ╞══════╪══════╪══════╪══════╪═════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    ┆ 1   │
    │ 2    ┆ 2    ┆ 2    ┆ b    ┆ 2   │
    │ 3    ┆ 3    ┆ 3    ┆ c    ┆ 3   │
    │ 4    ┆ 4    ┆ 4    ┆ d    ┆ 4   │
    │ 5    ┆ 5    ┆ 5    ┆ e    ┆ 5   │
    └──────┴──────┴──────┴──────┴─────┘

    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Copying column {self._in_col!r} to {self._out_col!r} ...")
        return frame.with_columns(pl.col(self._in_col).alias(self._out_col))


class CopyColumnsTransformer(BaseInNTransformer):
    r"""Implement a transformer to copy some columns.

    Args:
        columns: The columns to copy. ``None`` means all the
            columns.
        prefix: The column name prefix for the copied columns.
        suffix: The column name suffix for the copied columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import CopyColumns
    >>> transformer = CopyColumns(columns=["col1", "col3"], prefix="", suffix="_raw")
    >>> transformer
    CopyColumnsTransformer(columns=('col1', 'col3'), exclude_columns=(), missing_policy='raise', exist_policy='raise', prefix='', suffix='_raw')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_raw ┆ col3_raw │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ str  ┆ str  ┆ i64      ┆ str      │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    ┆ 1        ┆ 1        │
    │ 2    ┆ 2    ┆ 2    ┆ b    ┆ 2        ┆ 2        │
    │ 3    ┆ 3    ┆ 3    ┆ c    ┆ 3        ┆ 3        │
    │ 4    ┆ 4    ┆ 4    ┆ d    ┆ 4        ┆ 4        │
    │ 5    ┆ 5    ┆ 5    ┆ e    ┆ 5        ┆ 5        │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        prefix: str,
        suffix: str,
        exclude_columns: Sequence[str] = (),
        exist_policy: str = "raise",
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._prefix = prefix
        self._suffix = suffix

        check_column_exist_policy(exist_policy)
        self._exist_policy = exist_policy

    def get_args(self) -> dict:
        return super().get_args() | {
            "exist_policy": self._exist_policy,
            "prefix": self._prefix,
            "suffix": self._suffix,
        }

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self._check_output_columns(frame)
        logger.info(
            f"Copying {len(self.find_columns(frame)):,} columns | prefix={self._prefix!r} | "
            f"suffix={self._suffix!r} ..."
        )
        columns = self.find_common_columns(frame)
        out = frame.with_columns(
            frame.select(pl.col(columns)).rename(lambda name: f"{self._prefix}{name}{self._suffix}")
        )
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return out

    def _check_output_columns(self, frame: pl.DataFrame) -> None:
        r"""Check if the output columns already exist.

        Args:
            frame: The input DataFrame to check.
        """
        check_existing_columns(
            frame,
            columns=[f"{self._prefix}{col}{self._suffix}" for col in self.find_columns(frame)],
            exist_policy=self._exist_policy,
        )



r"""Contain utility functions to configure the standard logging
library."""

from __future__ import annotations

__all__ = ["configure_logging"]

import logging

from grizz.utils.imports import is_colorlog_available

if is_colorlog_available():  # pragma: no cover
    import colorlog

logger = logging.getLogger(__name__)


def configure_logging(level: int = logging.INFO) -> None:
    r"""Configure the logging module with a colored formatter.

    Args:
        level: The lower level.
    """
    if not is_colorlog_available():
        logging.basicConfig(level=level)
        return

    handler = colorlog.StreamHandler()
    formatter = colorlog.ColoredFormatter(
        fmt=(
            "%(log_color)s(%(process)d) %(asctime)s [%(levelname)s] %(name)s:%(lineno)s%(reset)s "
            "%(message_log_color)s%(message)s"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "bold_yellow",
            "ERROR": "red",
            "CRITICAL": "bold_red",
        },
        secondary_log_colors={
            "message": {
                "DEBUG": "cyan",
                "INFO": "reset",
                "WARNING": "bold_yellow",
                "ERROR": "red",
                "CRITICAL": "bold_red",
            }
        },
    )
    handler.setFormatter(formatter)

    logging.basicConfig(level=level, handlers=[handler])



r"""Contain ``polars.DataFrame`` transformers to convert some columns to
a new data type."""

from __future__ import annotations

__all__ = ["JsonDecodeTransformer"]

import logging
from typing import TYPE_CHECKING, Any, Union

import polars as pl
import polars.selectors as cs

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.column import check_column_exist_policy, check_existing_columns

if TYPE_CHECKING:
    from collections.abc import Sequence

    from polars.type_aliases import PythonDataType

logger = logging.getLogger(__name__)

PolarsDataType = Union[pl.DataType, type[pl.DataType]]


class JsonDecodeTransformer(BaseInNTransformer):
    r"""Implement a transformer to parse string values as JSON.

    Args:
        columns: The columns to parse. ``None`` means all the
            columns.
        prefix: The column name prefix for the copied columns.
        suffix: The column name suffix for the copied columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        dtype: The dtype to cast the extracted value to.
            If ``None``, the dtype will be inferred from the JSON
            value.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import JsonDecode
    >>> transformer = JsonDecode(columns=["col1", "col3"], prefix="", suffix="_out")
    >>> transformer
    JsonDecodeTransformer(columns=('col1', 'col3'), exclude_columns=(), missing_policy='raise', exist_policy='raise', prefix='', suffix='_out', dtype=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["[1, 2]", "[2]", "[1, 2, 3]", "[4, 5]", "[5, 4]"],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["['1', '2']", "['2']", "['1', '2', '3']", "['4', '5']", "['5', '4']"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌───────────┬──────┬─────────────────┬──────┐
    │ col1      ┆ col2 ┆ col3            ┆ col4 │
    │ ---       ┆ ---  ┆ ---             ┆ ---  │
    │ str       ┆ str  ┆ str             ┆ str  │
    ╞═══════════╪══════╪═════════════════╪══════╡
    │ [1, 2]    ┆ 1    ┆ ['1', '2']      ┆ a    │
    │ [2]       ┆ 2    ┆ ['2']           ┆ b    │
    │ [1, 2, 3] ┆ 3    ┆ ['1', '2', '3'] ┆ c    │
    │ [4, 5]    ┆ 4    ┆ ['4', '5']      ┆ d    │
    │ [5, 4]    ┆ 5    ┆ ['5', '4']      ┆ e    │
    └───────────┴──────┴─────────────────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌───────────┬──────┬─────────────────┬──────┐
    │ col1      ┆ col2 ┆ col3            ┆ col4 │
    │ ---       ┆ ---  ┆ ---             ┆ ---  │
    │ list[i64] ┆ str  ┆ list[str]       ┆ str  │
    ╞═══════════╪══════╪═════════════════╪══════╡
    │ [1, 2]    ┆ 1    ┆ ["1", "2"]      ┆ a    │
    │ [2]       ┆ 2    ┆ ["2"]           ┆ b    │
    │ [1, 2, 3] ┆ 3    ┆ ["1", "2", "3"] ┆ c    │
    │ [4, 5]    ┆ 4    ┆ ["4", "5"]      ┆ d    │
    │ [5, 4]    ┆ 5    ┆ ["5", "4"]      ┆ e    │
    └───────────┴──────┴─────────────────┴──────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        prefix: str,
        suffix: str,
        exclude_columns: Sequence[str] = (),
        exist_policy: str = "raise",
        missing_policy: str = "raise",
        dtype: PolarsDataType | PythonDataType | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._prefix = prefix
        self._suffix = suffix
        check_column_exist_policy(exist_policy)
        self._exist_policy = exist_policy

        self._dtype = dtype
        self._kwargs = kwargs

    def get_args(self) -> dict:
        return (
            super().get_args()
            | {
                "exist_policy": self._exist_policy,
                "prefix": self._prefix,
                "suffix": self._suffix,
                "dtype": self._dtype,
            }
            | self._kwargs
        )

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self._check_output_columns(frame)
        logger.info(f"Converting {len(self.find_columns(frame)):,} columns to JSON...")
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select(
                (cs.by_name(columns) & cs.string())
                .str.replace_all("'", '"')
                .str.json_decode(self._dtype, **self._kwargs)
            ).rename(lambda col: f"{self._prefix}{col}{self._suffix}")
        )

    def _check_output_columns(self, frame: pl.DataFrame) -> None:
        r"""Check if the output columns already exist.

        Args:
            frame: The input DataFrame to check.
        """
        check_existing_columns(
            frame,
            columns=[f"{self._prefix}{col}{self._suffix}" for col in self.find_columns(frame)],
            exist_policy=self._exist_policy,
        )
